/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.fortiss.tooling.base.ui.editor.fx;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.rectangular.RectangularResizableContentControllerBase;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.IModelChangeListener;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.utils.LayoutModelElementAdapter;

import javafx.scene.Node;
import javafx.scene.control.MenuItem;

/**
 * {@link RectangularResizableContentControllerBase} with a listener mechanism connected to
 * {@link EObject#eAdapters()}.
 */
public abstract class EObjectBasedRectangularResizableContentControllerBase
		extends RectangularResizableContentControllerBase implements IModelChangeProvider {
	/** Layout model adapter used for listening to EMF model changes. */
	private final LayoutModelElementAdapter layoutAdapter;

	/** Constructor. */
	public EObjectBasedRectangularResizableContentControllerBase(IContentMVCBundle mvcb) {
		super(mvcb);
		if(!(mvcb.getModel() instanceof ILayoutedModelElement)) {
			throw new IllegalArgumentException("Model element must be an instance of an EObject.");
		}
		ILayoutedModelElement lme = (ILayoutedModelElement)mvcb.getModel();
		layoutAdapter = new LayoutModelElementAdapter(lme) {
			@Override
			public void layoutChanged(Notification notification) {
				fireUpdateComplete();
			}
		};
	}

	/** Returns the {@link ILayoutedModelElement} model element. */
	protected ILayoutedModelElement getLayoutedME() {
		// wild cast works: see constructor check with exception
		return (ILayoutedModelElement)getModel();
	}

	// /** {@inheritDoc} */
	// @Override
	// protected void link(IMVCBundle startBundle, DiagramCoordinate startLocation,
	// IMVCBundle endBundle, DiagramCoordinate endLocation) {
	// createLinkByCommand(startBundle, startLocation, endBundle, endLocation,
	// (EObject)getViewer().getRootElement(), context);
	// }
	//
	// /** {@inheritDoc} */
	// @Override
	// public boolean isPossibleLinkTarget() {
	// return true;
	// }

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		// // outer-most grid areas of the diagram are inaccessible
		// DiagramViewerFeatures features = getViewer().getFeatures();
		// double x = max(features.getHorizontalSpacing(), diagramLocation.getX());
		// double y = max(features.getVerticalSpacing(), diagramLocation.getY());
		// // wild cast works: see constructor exception
		// EObject modelParent = (EObject)getModel();
		// IEclipseContext subContext =
		// createSubContextWithCompositionBorderLocation(context, modelParent, x, y);
		//
		// IContextMenuService icms = context.get(IContextMenuService.class);
		// return icms.createDefaultContextMenu(subContext);
		return Collections.emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public void delete() {
		// deleteByService(getEObject(), context);
	}

	/** The model change listener list. */
	private final List<IModelChangeListener> changeListeners =
			new LinkedList<IModelChangeListener>();

	/** {@inheritDoc} */
	@Override
	public void addListener(IModelChangeListener l) {
		if(changeListeners.isEmpty()) {
			connectToModel();
		}
		if(!changeListeners.contains(l)) {
			changeListeners.add(l);
		}
	}

	/** Called when sub-class should register with the underlying {@link EObject}. */
	protected void connectToModel() {
		getLayoutedME().eAdapters().add(layoutAdapter);
	}

	/** {@inheritDoc} */
	@Override
	public void removeListener(IModelChangeListener l) {
		changeListeners.remove(l);
		if(changeListeners.isEmpty()) {
			disconnectFromModel();
		}
	}

	/** Called when sub-class should unregister from the underlying {@link EObject}. */
	protected void disconnectFromModel() {
		getLayoutedME().eAdapters().remove(layoutAdapter);
	}

	/** {@inheritDoc} */
	@Override
	public void fireUpdateComplete() {
		EObject eo = getLayoutedME();
		for(IModelChangeListener mcl : changeListeners) {
			mcl.changed(eo);
		}
	}
}
