/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static java.util.Objects.requireNonNull;
import static org.fortiss.tooling.base.model.layout.LayoutPackage.ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.LayoutPackage;

/**
 * A specific {@link EContentAdapter} used for layouted model elements and layout data.
 * 
 * @author hoelzl
 */
public class LayoutModelElementAdapter extends EContentAdapter {
	/** The layouted model element. */
	private final ILayoutedModelElement modelElement;

	/** Constructor. */
	public LayoutModelElementAdapter(ILayoutedModelElement layouted) {
		this.modelElement = requireNonNull(layouted);
		modelElement.eAdapters().add(this);
	}

	/** Removes this adapter from the underlying model element. */
	public void unregisterAdapter() {
		modelElement.eAdapters().remove(this);
	}

	/** {@inheritDoc} */
	@Override
	public void notifyChanged(Notification notification) {
		Object notifier = notification.getNotifier();
		if(notifier == modelElement) {
			Object feature = notification.getFeature();
			if(feature instanceof EAttribute) {
				attributeChanged(notification);
				// no selfAdapt needed for attributes
				return;
			}
			if(feature == LayoutPackage.Literals.ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA) {
				super.notifyChanged(notification);
				layoutChanged(notification);
				return;
			}
		} else if(notifier instanceof ILayoutData) {
			super.notifyChanged(notification);
			layoutChanged(notification);
			return;
		}
		// prevent self-adapt for all other cases
	}

	/**
	 * Called when an attribute of the layouted model element receives the given notification.
	 * 
	 * @param notification
	 *            the notification
	 */
	protected void attributeChanged(Notification notification) {
		// default does nothing
	}

	/**
	 * Called when layout data of the layouted model element receives the given notification.
	 * 
	 * @param notification
	 *            the notification
	 */
	protected void layoutChanged(Notification notification) {
		// default does nothing
	}

	/** {@inheritDoc} */
	@Override
	protected void selfAdapt(Notification notification) {
		// ignore resources and resource sets
		if(notification.getNotifier() instanceof EObject) {
			super.selfAdapt(notification);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void handleContainment(Notification notification) {
		if(notification.getNotifier() instanceof ILayoutData) {
			super.handleContainment(notification);
			return;
		}

		Class<?> instanceClass = LayoutPackage.Literals.ILAYOUTED_MODEL_ELEMENT.getInstanceClass();
		if(notification.getFeatureID(instanceClass) == ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA) {
			super.handleContainment(notification);
		}
	}
}
