/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base;

import static java.util.Collections.emptyList;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.NEW_LINK;

import java.util.List;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.change.Change;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IClickController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.MVCBundleTag;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.impl.MVCBundlePartBase;

import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.control.MenuItem;
import javafx.scene.input.KeyCode;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;

/** Base class for {@link IController} implementations with move and link operation. */
public abstract class ControllerBase extends MVCBundlePartBase implements IController {
	/** Constructor. */
	public ControllerBase(IMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnShiftMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		return new ClickControllerBase() {
			/** {@inheritDoc} */
			@Override
			public void singleClick(Node node, DiagramCoordinate diagramLocation,
					boolean ctrlKeyDown, boolean shiftKeyDown) {
				if(!selectOnClick()) {
					return;
				}
				if(shiftKeyDown) {
					getViewer().handleShiftSelectionOf(getMVCBundle());
				} else {
					getViewer().setSingleSelectedMVCBundle(getMVCBundle());
				}
			}

			/** {@inheritDoc} */
			@Override
			public void secondaryClick(Node node, DiagramCoordinate diagramLocation,
					boolean ctrlKeyDown, boolean shiftKeyDown) {
				DiagramViewer viewer = getViewer();
				if(selectOnClick()) {
					viewer.setSingleSelectedMVCBundle(getMVCBundle());
				}
				if(ctrlKeyDown) {
					viewer.getModelModifier().accept(new Change() {
						@Override
						public void applyChange() {
							delete();
						}
					});
				} else {
					viewer.showContextMenu(node, diagramLocation, getMVCBundle());
				}
			}
		};
	}

	/** Returns whether the visual should be selected on single click. */
	protected boolean selectOnClick() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(allowLink() && gesture == NEW_LINK) {
			return getLinkCreationController();
		}
		return null;
	}

	/** Returns whether this visual can be linked. */
	protected boolean allowLink() {
		return true;
	}

	/** Returns the default {@link IDragController} for creating links. */
	protected IDragController getLinkCreationController() {
		return defaultLinkController;
	}

	/** The default {@link IDragController} for link creation. */
	private final IDragController defaultLinkController = new DragControllerBase() {
		@Override
		public void safeDragStarted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			diagramLocation = getVisual().getLinkAnchorage(diagramLocation);
			getViewer().startNewLinkLineFeedback(bundle, node, diagramLocation);
		}

		@Override
		protected void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			diagramLocation = bundle.getVisual().getLinkAnchorage(diagramLocation);
			getViewer().updateNewLinkLineFeedback(diagramLocation,
					newLinkFeedback(getMVCBundle(), startDiagramLocation, bundle, diagramLocation));
		}

		@Override
		protected void safeDragCompleted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			getViewer().terminateNewLinkLineFeedback();
			getViewer().getModelModifier().accept(new Change() {
				@Override
				public void applyChange() {
					link(getMVCBundle(), startDiagramLocation, bundle, diagramLocation);
				}
			});
		}
	};

	/**
	 * Creates a link between the bundles at the given, respective locations.
	 * 
	 * @param startBundle
	 *            the {@link IMVCBundle} of the start node of the link
	 * @param startLocation
	 *            the location of the start of the link
	 * @param endBundle
	 *            the {@link IMVCBundle} of the end node of the link
	 * @param endLocation
	 *            the location of the end of the link
	 */
	protected void link(IMVCBundle startBundle, DiagramCoordinate startLocation,
			IMVCBundle endBundle, DiagramCoordinate endLocation) {
		// the default does nothing
	}

	/**
	 * Determines the feedback color of for creating new links between the given bundles.
	 * 
	 * @param startBundle
	 *            the {@link IMVCBundle} of the start node of the link
	 * @param startLocation
	 *            the location of the start of the link
	 * @param endBundle
	 *            the {@link IMVCBundle} of the end node of the link
	 * @param endLocation
	 *            the location of the end of the link
	 * @return the feedback line color
	 */
	protected Paint newLinkFeedback(IMVCBundle startBundle, DiagramCoordinate startLocation,
			IMVCBundle endBundle, DiagramCoordinate endLocation) {
		return Color.BLACK;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isPossibleLinkTarget() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForNewLinkFrom(IMVCBundle linkStartBundle) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForMoveLink(ILinkMVCBundle linkBundle,
			IAnchorageMVCBundle movedAnchorageBundle) {
		return null;
	}

	/** Delete the model of this bundle. */
	protected void delete() {
		// the default does nothing
	}

	/**
	 * The default implementation for key press controller allowing move with cursor keys and delete
	 * with the delete key.
	 */
	private final IKeyPressController defaultKeyPressController = new IKeyPressController() {
		@Override
		public void keyTyped(KeyCode code) {
			DiagramViewer viewer = getViewer();
			if(code == KeyCode.DELETE) {
				viewer.getModelModifier().accept(new Change() {
					@Override
					public void applyChange() {
						delete();
					}
				});
			}
		}
	};

	/** {@inheritDoc} */
	@Override
	public IKeyPressController getKeyPressController(Node node) {
		return defaultKeyPressController;
	}

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		switch(edg) {
			case NEW_LINK:
				return allowLink() ? Cursor.CROSSHAIR : Cursor.DEFAULT;
			default:
				return Cursor.DEFAULT;
		}
	}
}
