/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.scene.Node;

/** Base implementation of {@link IDragController}. */
public abstract class DragControllerBase implements IDragController {
	/** Stores the start bundle of the drag gesture. */
	protected IMVCBundle startBundle = null;
	/** Stores the start node of the drag gesture. */
	protected Node startNode;
	/** Stores the start location of the drag gesturein diagram coordinates. */
	protected DiagramCoordinate startDiagramLocation = null;

	/** {@inheritDoc} */
	@Override
	public final void dragStarted(IMVCBundle bundle, Node node, DiagramCoordinate diagramLocation) {
		this.startBundle = bundle;
		this.startNode = node;
		this.startDiagramLocation = diagramLocation;
		safeDragStarted(bundle, node, startDiagramLocation);
	}

	/**
	 * Informs the drag controller that a drag has started.
	 * 
	 * @param bundle
	 *            the final {@link IMVCBundle}
	 * @param node
	 *            the final node
	 * @param diagramLocation
	 *            the current absolute mouse location
	 */
	protected void safeDragStarted(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// the default does nothing
	}

	/** {@inheritDoc} */
	@Override
	public final void dragInProgress(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// filter drag events that arrive before drag started was called.
		if(isDragInProgress()) {
			safeDragInProgress(bundle, node, diagramLocation);
		}
	}

	/**
	 * Updates the drag controller with the new drag location guaranteeing that its only called
	 * after {@link #dragStarted(IMVCBundle, Node, DiagramCoordinate)} was called and before
	 * {@link #dragCompleted(IMVCBundle, Node, DiagramCoordinate)} is
	 * called.
	 * 
	 * @param bundle
	 *            the current {@link IMVCBundle}
	 * @param node
	 *            the current node under the drag gesture pointer
	 * @param diagramLocation
	 *            the current diagram location of the drag gesture
	 */
	protected void safeDragInProgress(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// the default does nothing
	}

	/** {@inheritDoc} */
	@Override
	public final void dragCompleted(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		if(isDragInProgress()) {
			safeDragCompleted(bundle, node, diagramLocation);
		}
		startBundle = null;
		startNode = null;
		startDiagramLocation = null;
	}

	/**
	 * Updates the drag controller with the final drag location guaranteeing that its only called
	 * after {@link #dragStarted(IMVCBundle, Node, DiagramCoordinate)} was called and any number of
	 * {@link #safeDragInProgress(IMVCBundle, Node, DiagramCoordinate)} occurred.
	 * 
	 * @param bundle
	 *            the final {@link IMVCBundle}
	 * @param node
	 *            the final node under the gesture pointer
	 * @param diagramLocation
	 *            the final diagram location of the drag gesture
	 */
	protected void safeDragCompleted(IMVCBundle bundle, Node node,
			DiagramCoordinate diagramLocation) {
		// the default does nothing
	}

	/**
	 * Returns whether a drag gesture is in progress, i.e., dragStarted() was called and
	 * dragCompleted() was not yet called.
	 */
	protected final boolean isDragInProgress() {
		return startBundle != null;
	}
}
