/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base;

import static java.util.Collections.emptyList;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange.locationFeedbackChange;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base.DraggingUtils.clampAndSnapMoveLocationToGridCenterPoint;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.ILinkVisual.END_OF_LINK_BEND_POINT_INDEX;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.ILinkVisual.NO_BEND_POINT_INDEX;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.ILinkVisual.SPECIAL_BEND_POINT_INDEX_CAP;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.ILinkVisual.START_OF_LINK_BEND_POINT_INDEX;

import java.util.List;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IClickController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.MVCBundleTag;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.impl.MVCBundlePartBase;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.ILinkVisual;

import javafx.geometry.Point2D;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.control.MenuItem;
import javafx.scene.input.KeyCode;

/**
 * Base class implementing {@link IController} for links. It provides dragging implementation for
 * visuals of bend-points and of the start or end of the link. It also handles creation and removal
 * of bend-points (including keyboard interaction).
 * <P>
 * Sub-classes must implement the abstract methods, which are called once the respective operation
 * is completed, and write the results back into the underlying data model.
 */
public abstract class LinkControllerBase extends MVCBundlePartBase implements IController {
	/** Stores the index of the currently moving bend-point. */
	private int currentBendPointIndex = 0;

	/** Constructor. */
	public LinkControllerBase(ILinkMVCBundle mvcb) {
		super(mvcb);
	}

	/** Create a new bend point with the given mouse click coordinates. */
	protected abstract void createBendPointAt(int bpIndex, DiagramCoordinate location);

	/**
	 * Moves the model element.
	 * 
	 * @param bpIndex
	 *            the index of moved bend-point
	 * @param dx
	 *            the x delta
	 * @param dy
	 *            the y delta
	 */
	protected abstract void updateModelAfterBendPointMove(int bpIndex, double dx, double dy);

	/** Deletes the current bend-point. */
	protected abstract void deleteBendPoint(int bpIndex);

	/**
	 * Moves the given end of the link to the new anchorage. The given bend-point
	 */
	protected abstract void updateModelAfterLinkMove(ILinkMVCBundle linkBundle,
			IMVCBundle startBundle, IMVCBundle endBundle);

	/** Removes the given bend point. */
	protected abstract void deleteLink();

	/** Sets the current bend-point index. */
	protected final void setCurrentBendPointIndex(int bendPointIndex) {
		currentBendPointIndex = bendPointIndex;
	}

	/** {@inheritDoc} */
	@Override
	public ILinkMVCBundle getMVCBundle() {
		// wild cast works: see constructor
		return (ILinkMVCBundle)super.getMVCBundle();
	}

	/** Returns the link visual. */
	protected ILinkVisual getLinkVisual() {
		return getMVCBundle().getVisual();
	}

	/** Returns the {@link IClickController} for creating bend-points and selecting the link. */
	protected IClickController getBendPointCreateAndLinkSelectController() {
		return new ClickControllerBase() {
			@Override
			public void singleClick(Node node, DiagramCoordinate diagramLocation,
					boolean ctrlKeyDown, boolean shiftKeyDown) {
				if(ctrlKeyDown) {
					// link must be primary selected
					if(!getViewer().isPrimarySelected(getMVCBundle())) {
						return;
					}
					// control+click on visible or selection line creates a bend point
					DiagramViewerFeatures features = getViewer().getFeatures();
					diagramLocation = features.snapToCenter(diagramLocation);
					createBendPointAt(currentBendPointIndex, diagramLocation);
					getViewer().updateVisual(getMVCBundle());
				} else {
					if(shiftKeyDown) {
						getViewer().handleShiftSelectionOf(getMVCBundle());
					} else {
						getViewer().setSingleSelectedMVCBundle(getMVCBundle());
					}
				}
			}

			@Override
			public void secondaryClick(Node node, DiagramCoordinate locationOnNode,
					boolean ctrlKeyDown, boolean shiftKeyDown) {
				if(ctrlKeyDown) {
					// link must be primary selected
					if(!getViewer().isPrimarySelected(getMVCBundle())) {
						return;
					}
					// control+right-click deletes the link
					deleteLink();
					getViewer().updateFromModel();
				}
			}
		};
	}

	/** Returns the {@link IDragController} for moving bend-points. */
	protected IDragController getBendPointMoveController() {
		return new DragControllerBase() {
			@Override
			public void safeDragInProgress(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				Point2D delta = diagramLocation.subtract(startDiagramLocation);
				moveBendPointVisually(currentBendPointIndex, delta);
				getViewer().updateVisual(getMVCBundle());
			}

			@Override
			public void safeDragCompleted(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				Point2D delta = diagramLocation.subtract(startDiagramLocation);
				moveBendPointInModel(currentBendPointIndex, delta);
				getViewer().updateVisual(getMVCBundle());
			}
		};
	}

	/**
	 * Updates the bend-point visual(s) during the dragging phase of the move operation. Sub-classes
	 * may override to move other bend-points than only the current bend-point.
	 */
	protected void moveBendPointVisually(int bpIndex, Point2D delta) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, delta);
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, chg);
	}

	/** Returns the clamped and snapped move delta feedback change for the given bend point. */
	protected final FeedbackChange getMoveDeltaFeedback(int bpIndex, Point2D delta) {
		DiagramCoordinate bpModelLocation = getLinkVisual().getBendPointLocation(bpIndex);
		FeedbackChange chg = clampAndSnapMoveLocationToGridCenterPoint(bpModelLocation.getX(),
				bpModelLocation.getY(), delta.getX(), delta.getY(), getViewer().getFeatures());
		return chg;
	}

	/**
	 * Updates the bend-point model element(s) after the dragging of the move operation is complete.
	 */
	protected void moveBendPointInModel(int bpIndex, Point2D delta) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, delta);
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, null);
		updateModelAfterBendPointMove(bpIndex, chg.getDeltaX(), chg.getDeltaY());
	}

	/** Returns the {@link IClickController} for deleting bend-points. */
	protected IClickController getBendPointDeleteController() {
		return new ClickControllerBase() {
			@Override
			public void secondaryClick(Node node, DiagramCoordinate diagramLocation,
					boolean ctrlKeyDown, boolean shiftKeyDown) {
				if(ctrlKeyDown) {
					// bundle must be primary selection
					if(!getViewer().isPrimarySelected(getMVCBundle())) {
						return;
					}
					// control+right-click on selection handle of bend point removes it
					deleteBendPoint(currentBendPointIndex);
					getViewer().updateVisual(getMVCBundle());
				}
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(getViewer().getSelection().getPrimarySelection() == getMVCBundle()) {
			ILinkVisual linkVisual = getLinkVisual();
			int bpIndex = linkVisual.getBendPointIndex(node);
			if(bpIndex == START_OF_LINK_BEND_POINT_INDEX ||
					bpIndex == END_OF_LINK_BEND_POINT_INDEX) {
				return getLinkMoveController(bpIndex);
			}
			if(bpIndex == NO_BEND_POINT_INDEX) {
				return null;
			}
			setCurrentBendPointIndex(bpIndex);
			return getBendPointMoveController();
		}
		return null;
	}

	/**
	 * Returns an {@link IDragController} for relocating the end of this link. {@code bpIndex} is
	 * either {@link ILinkVisual#START_OF_LINK_BEND_POINT_INDEX} or
	 * {@link ILinkVisual#END_OF_LINK_BEND_POINT_INDEX} depending on which part of the link is
	 * moved.
	 */
	protected IDragController getLinkMoveController(final int bpIndex) {
		return new DragControllerBase() {
			@Override
			protected void safeDragStarted(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				ILinkMVCBundle linkBundle = getMVCBundle();
				IAnchorageMVCBundle anchorage;
				if(bpIndex == START_OF_LINK_BEND_POINT_INDEX) {
					anchorage = linkBundle.getStartAnchorage();
				} else {
					anchorage = linkBundle.getEndAnchorage();
				}
				getLinkVisual().disableFeedback();
				getViewer().startMoveLinkFeedback(linkBundle, anchorage);
			}

			@Override
			public void safeDragInProgress(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				Point2D linkAnchorageLocation =
						bundle.getVisual().getLinkAnchorage(diagramLocation);
				Point2D delta;
				if(linkAnchorageLocation != null) {
					delta = linkAnchorageLocation.subtract(startDiagramLocation);
				} else {
					delta = diagramLocation.subtract(startDiagramLocation);
				}
				FeedbackChange chg = locationFeedbackChange(delta.getX(), delta.getY());
				getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, chg);
				getViewer().updateVisual(getMVCBundle());
			}

			@Override
			public void safeDragCompleted(IMVCBundle bundle, Node node,
					DiagramCoordinate diagramLocation) {
				getLinkVisual().enableFeedback();
				getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, null);
				getViewer().terminateMoveLinkFeedback();
				ILinkMVCBundle linkBundle = getMVCBundle();
				if(bpIndex == START_OF_LINK_BEND_POINT_INDEX) {
					updateModelAfterLinkMove(linkBundle, bundle, linkBundle.getEndAnchorage());
				} else {
					updateModelAfterLinkMove(linkBundle, linkBundle.getStartAnchorage(), bundle);
				}
				getViewer().updateVisual(linkBundle);
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		int bpIndex = getLinkVisual().getBendPointIndex(node);
		if(bpIndex < SPECIAL_BEND_POINT_INDEX_CAP) {
			return null;
		}
		setCurrentBendPointIndex(bpIndex);
		if(getLinkVisual().isBendPointHandle(node)) {
			return getBendPointDeleteController();
		}
		return getBendPointCreateAndLinkSelectController();
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean selectOnShiftMousePress() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IKeyPressController getKeyPressController(Node node) {
		return new IKeyPressController() {
			@Override
			public void keyTyped(KeyCode code) {
				if(code == KeyCode.DELETE) {
					deleteLink();
				}
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isPossibleLinkTarget() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForNewLinkFrom(IMVCBundle linkStartBundle) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public MVCBundleTag getLinkTargetEffectForMoveLink(ILinkMVCBundle linkBundle,
			IAnchorageMVCBundle movedAnchorageBundle) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		switch(edg) {
			case MOVE_BENDPOINT:
				return allowBendPointMove() ? Cursor.MOVE : Cursor.DEFAULT;
			case NEW_BENDPOINT:
				return Cursor.CROSSHAIR;
			case MOVE_LINK:
				return allowLinkMove() ? Cursor.MOVE : Cursor.DEFAULT;
			default:
				return Cursor.DEFAULT;
		}
	}

	/** Determines whether bend points can be moved. */
	protected boolean allowBendPointMove() {
		return true;
	}

	/** Determines whether link start and end can be relocated. */
	protected boolean allowLinkMove() {
		return true;
	}
}
