/*******************************************************************************
 * Copyright (c) 2019 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base;

import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.MOVE;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange.locationFeedbackChange;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base.DraggingUtils.clampAndSnapMoveLocationToGridPoint;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerSelection;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.change.Change;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.change.ChangeSet;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundleWithParent;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.geometry.Rectangle2D;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.input.KeyCode;

/**
 * Controller implementation, which allows to move the underlying model element. Sub-classes must
 * implement {@link #move(FeedbackChange)} in order to change the underlying model with the
 * respective coordinate changes.
 */
public abstract class MoveControllerBase extends ControllerBase {
	/** Constructor. */
	public MoveControllerBase(IMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(allowMove() && gesture == MOVE) {
			return moveController;
		}
		return super.getDragController(gesture, node, diagramLocation);
	}

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		if(edg == EDragGesture.MOVE) {
			return allowMove() ? Cursor.MOVE : Cursor.DEFAULT;
		}
		return super.getCurrentCursor(node, diagramLocation);
	}

	/** {@inheritDoc} */
	@Override
	public IKeyPressController getKeyPressController(final Node node) {
		if(allowMove() && allowKeyboardMove()) {
			return new IKeyPressController() {
				@Override
				public void keyTyped(final KeyCode code) {
					DiagramViewerFeatures features = getViewer().getFeatures();
					double hSpacing = features.getHorizontalSpacing();
					double vSpacing = features.getVerticalSpacing();
					FeedbackChange fbChange = null;
					if(code == KeyCode.RIGHT) {
						fbChange = locationFeedbackChange(hSpacing, 0);
					} else if(code == KeyCode.LEFT) {
						fbChange = locationFeedbackChange(-hSpacing, 0);
					} else if(code == KeyCode.UP) {
						fbChange = locationFeedbackChange(0, -vSpacing);
					} else if(code == KeyCode.DOWN) {
						fbChange = locationFeedbackChange(0, vSpacing);
					} else {
						MoveControllerBase.super.getKeyPressController(node).keyTyped(code);
					}
					if(fbChange != null) {
						handleMoveCompleted(fbChange);
					}
				}
			};
		}
		return super.getKeyPressController(node);
	}

	/**
	 * Perform the move operation by altering the underlying model. The default does nothing.
	 * 
	 * @param deltaChange
	 *            the (possibly altered) delta the model should be moved.
	 */
	protected void move(FeedbackChange deltaChange) {
		// default does nothing
	}

	/** Returns whether the model element can be moved. */
	protected boolean allowMove() {
		return true;
	}

	/** Returns whether the model element can be moved by keyboard commands. */
	protected boolean allowKeyboardMove() {
		return true;
	}

	/**
	 * Returns whether move operations should be clamped to positive diagram area and snapped to the
	 * grid.
	 */
	protected boolean clampAndSnapMove() {
		return true;
	}

	/**
	 * Alters the move feedback to apply some constraints to it (e.g. no negative result location).
	 */
	protected FeedbackChange alterMoveFeedback(FeedbackChange change) {
		if(!clampAndSnapMove()) {
			return change;
		}
		Rectangle2D modelBounds = getVisual().getModelBounds();
		return clampAndSnapMoveLocationToGridPoint(modelBounds.getMinX(), modelBounds.getMinY(),
				change.getDeltaX(), change.getDeltaY(), getViewer().getFeatures());
	}

	/** Performs an update of the feedback changes of the currently selected model elements. */
	private void handleMoveDragInProgress(FeedbackChange delta) {
		for(MoveControllerBase ctrl : getMoveControllerList()) {
			FeedbackChange altered = ctrl.alterMoveFeedback(delta);
			ctrl.getMVCBundle().setFeedbackChange(altered);
		}
	}

	/** Completes the move operation of the currently selected model elements. */
	private void handleMoveCompleted(FeedbackChange delta) {
		ChangeSet changeSet = new ChangeSet();
		for(final MoveControllerBase ctrl : getMoveControllerList()) {
			ctrl.getMVCBundle().setFeedbackChange(FeedbackChange.EMPTY);
			final FeedbackChange altered = ctrl.alterMoveFeedback(delta);
			if(altered.isEmpty()) {
				continue;
			}
			changeSet.add(new Change() {
				@Override
				public void applyChange() {
					ctrl.move(altered);
				}
			});
		}
		if(!changeSet.isEmpty()) {
			getViewer().getModelModifier().accept(changeSet);
		}
	}

	/** Returns a list of {@link MoveControllerBase}s derived from the current diagram selection. */
	private List<MoveControllerBase> getMoveControllerList() {
		DiagramViewer viewer = getViewer();
		DiagramViewerSelection selection = viewer.getSelection();
		List<MoveControllerBase> controllers = new ArrayList<MoveControllerBase>();
		IMVCBundle primary = selection.getPrimarySelection();
		IController primaryController = primary.getController();
		if(primaryController instanceof MoveControllerBase &&
				((MoveControllerBase)primaryController).allowMove()) {
			controllers.add((MoveControllerBase)primaryController);
		}
		if(primary instanceof IContentMVCBundle) {
			addChildMoveControllers(((IContentMVCBundle)primary).getChildren(), controllers);
		}
		for(IMVCBundle secondary : selection.getSecondarySelections()) {
			viewer.updateVisual(secondary);
			IController secondaryController = secondary.getController();
			if(secondaryController instanceof MoveControllerBase &&
					((MoveControllerBase)secondaryController).allowMove()) {
				controllers.add((MoveControllerBase)secondaryController);
			}
			if(secondary instanceof IContentMVCBundle) {
				addChildMoveControllers(((IContentMVCBundle)secondary).getChildren(), controllers);
			}
		}
		return controllers;
	}

	/** Recursively adds the children controllers to the given result list. */
	private void addChildMoveControllers(List<IContentMVCBundleWithParent> children,
			List<MoveControllerBase> result) {
		for(IContentMVCBundleWithParent child : children) {
			if(child.getController() instanceof MoveControllerBase) {
				result.add((MoveControllerBase)child.getController());
			}
			addChildMoveControllers(((IContentMVCBundle)child).getChildren(), result);
		}
	}

	/** The drag controller managing the move operation. */
	private final IDragController moveController = new DragControllerBase() {
		@Override
		protected void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			handleMoveDragInProgress(getMoveDelta(diagramLocation));
		}

		@Override
		protected void safeDragCompleted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			handleMoveCompleted(getMoveDelta(diagramLocation));
		}

		/** Returns the move delta. */
		private FeedbackChange getMoveDelta(DiagramCoordinate endDiagramLocation) {
			double dx = endDiagramLocation.getX() - startDiagramLocation.getX();
			double dy = endDiagramLocation.getY() - startDiagramLocation.getY();
			return locationFeedbackChange(dx, dy);
		}
	};
}
