/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.rectangular;

import static java.lang.Math.hypot;
import static java.util.Objects.requireNonNull;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;

/**
 * This class encapsulates locations on the border of a rectangle defined by the side and offset
 * value. It can be used to compute x-y-locations relative to the upper-left corner of the
 * rectangle.
 * <P>
 * The {@link #getClosestLocationOnBounds(Point2D, Rectangle2D, Dimension2D)} method allows to
 * construct a {@link RectangularBorderLocation} given a rectangle and a location. The computed
 * location will be on the rectangle and it will be the closest of those locations to the given
 * point.
 */
public final class RectangularBorderLocation {
	/** The side value. */
	private final Side side;
	/** The offset value. */
	private final double offset;
	/** The rectangle used for the computation. */
	private final Rectangle2D rectangle;
	/**
	 * The correction values used to compute the final location (usually half the size of the object
	 * located on the border so that it looks centered.
	 */
	private final Dimension2D correction;

	/** Constructor. */
	public RectangularBorderLocation(Side side, double offset, Rectangle2D rectangle,
			Dimension2D correction) {
		this.side = requireNonNull(side);
		this.offset = offset;
		this.rectangle = requireNonNull(rectangle);
		this.correction = requireNonNull(correction);
	}

	/** Returns the side. */
	public Side getSide() {
		return side;
	}

	/** Returns the offset. */
	public double getOffset() {
		return offset;
	}

	/**
	 * Returns the location computed from this {@link RectangularBorderLocation} relative to the
	 * upper-left corner of the {@link #rectangle}.
	 */
	public DiagramCoordinate getLocation() {
		if(side == Side.TOP) {
			return new DiagramCoordinate(offset, -correction.getHeight() / 2);
		}
		if(side == Side.LEFT) {
			return new DiagramCoordinate(-correction.getWidth() / 2, offset);
		}
		if(side == Side.BOTTOM) {
			return new DiagramCoordinate(offset,
					rectangle.getHeight() - correction.getHeight() / 2);
		}
		// Side.RIGHT
		return new DiagramCoordinate(rectangle.getWidth() - correction.getWidth() / 2, offset);
	}

	/**
	 * Returns the closest point on the rectangle bounds when pointing at the given location. Only
	 * locations with distances {@code horizontalSpan} and {@code verticalSpan} are considered.
	 */
	public static final RectangularBorderLocation getClosestLocationOnBounds(Point2D p,
			Rectangle2D rectangle, Dimension2D span) {
		LocationOnBorderComputation comp = new LocationOnBorderComputation(p, rectangle, span);
		return comp.getResult();
	}

	/** Class for efficiently computing the closest viable position on the parent bounds. */
	private static class LocationOnBorderComputation {
		/** The left-right result. */
		private RectangularBorderLocation leftRight;
		/** The left-right distance. */
		private double leftRightDistance;

		/** The top-bottom result. */
		private RectangularBorderLocation topBottom;
		/** The top-bottom distance. */
		private double topBottomDistance;

		/**
		 * Constructor, which computes the closest location on the given border to the given point.
		 */
		public LocationOnBorderComputation(Point2D pt, Rectangle2D rectangle, Dimension2D span) {
			// divide the problem by reducing to horizontal and vertical sub-problems
			computeLeftRight(pt, rectangle, span);
			computeTopBottom(pt, rectangle, span);
		}

		/** Computes the minimal location on the top or bottom border. */
		private void computeTopBottom(Point2D pt, Rectangle2D rect, Dimension2D span) {
			double horizontalSpan = span.getWidth();
			double hs2 = horizontalSpan / 2;
			Side side = Side.TOP;
			double borderY = 0;
			if(pt.getY() > rect.getHeight() / 2) {
				side = Side.BOTTOM;
				borderY = rect.getHeight();
			}
			double xOffset = 0;
			// while the offset is zero, the hypot() calculation uses half the span as correction
			double currentDistance = hypot(pt.getX() - hs2, pt.getY() - borderY);
			for(double nextOffset = horizontalSpan; nextOffset < rect.getWidth(); nextOffset +=
					horizontalSpan) {
				double nextDistance = hypot(pt.getX() - hs2 - nextOffset, pt.getY() - borderY);
				if(nextDistance > currentDistance) {
					// distance increased => solution found
					break;
				}
				xOffset = nextOffset;
				currentDistance = nextDistance;
			}
			topBottom = new RectangularBorderLocation(side, xOffset, rect, span);
			topBottomDistance = currentDistance;
		}

		/** Computes the minimal location on the left or right border. */
		private void computeLeftRight(Point2D pt, Rectangle2D rect, Dimension2D span) {
			double verticalSpan = span.getHeight();
			double vs2 = verticalSpan / 2;
			Side side = Side.LEFT;
			double borderX = 0;
			if(pt.getX() > rect.getWidth() / 2) {
				side = Side.RIGHT;
				borderX = rect.getWidth();
			}
			double yOffset = 0;
			double currentDistance = hypot(pt.getX() - borderX, pt.getY() - vs2);
			for(double nextOffset = verticalSpan; nextOffset < rect.getHeight(); nextOffset +=
					verticalSpan) {
				double nextDistance = hypot(pt.getX() - borderX, pt.getY() - vs2 - nextOffset);
				if(nextDistance > currentDistance) {
					// distance increased => solution found
					break;
				}
				yOffset = nextOffset;
				currentDistance = nextDistance;
			}
			leftRight = new RectangularBorderLocation(side, yOffset, rect, span);
			leftRightDistance = currentDistance;
		}

		/** Returns the border location computation result. */
		public RectangularBorderLocation getResult() {
			// conquer the problem by comparing the sub-problem results
			if(leftRightDistance <= topBottomDistance) {
				return leftRight;
			}
			return topBottom;
		}
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return side.toString() + ":" + offset;
	}
}
